@@ -158,7 +158,7 @@ void SMRFilter::prepared(PointTableRef table)
                                              "NumberOfReturns. Skipping "
                                              "segmentation of last returns and "
                                              "proceeding with all returns.\n";
-            m_args->m_returns = {""};
+            m_args->m_returns.clear();
         }
     }
 }
