@@ -55,46 +55,6 @@ namespace
         "http://pdal.io/stages/reader.ept.html",
         { "ept" }
     };
-
-    Dimension::Type getType(const Json::Value& dim)
-    {
-        if (dim.isMember("scale"))
-            return Dimension::Type::Double;
-
-        const std::string type(dim["type"].asString());
-        const uint64_t size(dim["size"].asUInt64());
-
-        if (type == "signed")
-        {
-            switch (size)
-            {
-                case 1: return Dimension::Type::Signed8;
-                case 2: return Dimension::Type::Signed16;
-                case 4: return Dimension::Type::Signed32;
-                case 8: return Dimension::Type::Signed64;
-            }
-        }
-        else if (type == "unsigned")
-        {
-            switch (size)
-            {
-                case 1: return Dimension::Type::Unsigned8;
-                case 2: return Dimension::Type::Unsigned16;
-                case 4: return Dimension::Type::Unsigned32;
-                case 8: return Dimension::Type::Unsigned64;
-            }
-        }
-        else if (type == "float")
-        {
-            switch (size)
-            {
-                case 4: return Dimension::Type::Float;
-                case 8: return Dimension::Type::Double;
-            }
-        }
-
-        return Dimension::Type::None;
-    }
 }
 
 CREATE_STATIC_STAGE(EptReader, s_info);
@@ -105,6 +65,10 @@ EptReader::EptReader()
 EptReader::~EptReader()
 {}
 
+EptReader::Args::Args()
+    : m_addons(new Json::Value())
+{ }
+
 std::string EptReader::getName() const { return s_info.name; }
 
 void EptReader::addArgs(ProgramArgs& args)
@@ -112,6 +76,9 @@ void EptReader::addArgs(ProgramArgs& args)
     args.add("bounds", "Bounds to fetch", m_args.boundsArg());
     args.add("origin", "Origin of source file to fetch", m_args.originArg());
     args.add("threads", "Number of worker threads", m_args.threadsArg());
+    args.add("resolution", "Resolution limit", m_args.resolutionArg());
+    args.add("addons", "Mapping of addon dimensions to their output directory",
+            m_args.addonsArg());
 }
 
 BOX3D EptReader::Args::bounds() const
@@ -124,8 +91,8 @@ BOX3D EptReader::Args::bounds() const
     const BOX2D b(m_bounds.to2d());
     if (b.empty())
     {
-        double mn((std::numeric_limits<double>::lowest)());
-        double mx((std::numeric_limits<double>::max)());
+        const double mn((std::numeric_limits<double>::lowest)());
+        const double mx((std::numeric_limits<double>::max)());
         return BOX3D(mn, mn, mn, mx, mx, mx);
     }
 
@@ -140,32 +107,83 @@ void EptReader::initialize()
 {
     m_root = m_filename;
 
+    auto& debug(log()->get(LogLevel::Debug));
+
     const std::string prefix("ept://");
-    if (m_root.substr(0, prefix.size()) == prefix)
+    const std::string postfix("ept.json");
+    if (Utils::startsWith(m_root, prefix))
     {
         m_root = m_root.substr(prefix.size());
     }
+    if (Utils::endsWith(m_root, postfix))
+    {
+        m_root = m_root.substr(0, m_root.size() - postfix.size());
+    }
+
+    if (m_root.empty())
+        throwError("Missing input filename");
 
     m_arbiter.reset(new arbiter::Arbiter());
     m_ep.reset(new arbiter::Endpoint(m_arbiter->getEndpoint(m_root)));
-    m_pool.reset(new Pool(m_args.threads()));
-    log()->get(LogLevel::Debug) << "Endpoint: " << m_ep->prefixedRoot() <<
-        std::endl;
+    const std::size_t threads(m_args.threads());
+    if (threads > 100)
+    {
+        log()->get(LogLevel::Warning) << "Using a large thread count: " <<
+            threads << " threads" << std::endl;
+    }
+    m_pool.reset(new Pool(threads));
 
-    m_info.reset(new EptInfo(parse(m_ep->get("ept.json"))));
-    log()->get(LogLevel::Debug) << "Got EPT info" << std::endl;
+    debug << "Endpoint: " << m_ep->prefixedRoot() << std::endl;
+    try
+    {
+        m_info.reset(new EptInfo(parse(m_ep->get("ept.json"))));
+    }
+    catch (std::exception& e)
+    {
+        throwError(e.what());
+    }
 
-    // Set spatial reference if it's not overriden.
-    if (getSpatialReference().empty())
-        setSpatialReference(m_info->srs());
+    debug << "Got EPT info" << std::endl;
+    debug << "SRS: " << m_info->srs() << std::endl;
+    setSpatialReference(m_info->srs());
 
     // Figure out our query parameters.
     m_queryBounds = m_args.bounds();
-    handleOriginQuery();
+    try
+    {
+        handleOriginQuery();
+    }
+    catch (std::exception& e)
+    {
+        throwError(e.what());
+    }
 
-    log()->get(LogLevel::Debug) << "Query bounds: " << m_queryBounds <<
-        std::endl;
-    log()->get(LogLevel::Debug) << "Threads: " << m_pool->size() << std::endl;
+    // Figure out our max depth.
+    const double queryResolution(m_args.resolution());
+    if (queryResolution)
+    {
+        double currentResolution =
+            (m_info->bounds().maxx - m_info->bounds().minx) / m_info->span();
+
+        debug << "Root resolution: " << currentResolution << std::endl;
+
+        // To select the current resolution level, we need depthEnd to be one
+        // beyond it - this is a non-inclusive parameter.
+        ++m_depthEnd;
+
+        while (currentResolution > queryResolution)
+        {
+            currentResolution /= 2;
+            ++m_depthEnd;
+        }
+
+        debug << "Query resolution:  " << queryResolution << "\n";
+        debug << "Actual resolution: " << currentResolution << "\n";
+        debug << "Depth end: " << m_depthEnd << "\n";
+    }
+
+    debug << "Query bounds: " << m_queryBounds << "\n";
+    debug << "Threads: " << m_pool->size() << std::endl;
 }
 
 void EptReader::handleOriginQuery()
@@ -195,7 +213,7 @@ void EptReader::handleOriginQuery()
         // Otherwise it's a file path (or part of one - for example selecting
         // by a basename or a tile ID rather than a full path is convenient).
         // Find it within the sources list, and make sure it's specified
-        // uniquely enough to select only one filt.
+        // uniquely enough to select only one file.
         for (Json::ArrayIndex i(0); i < sources.size(); ++i)
         {
             const Json::Value& entry(sources[i]);
@@ -226,30 +244,43 @@ void EptReader::handleOriginQuery()
     const Json::Value& found(sources[static_cast<Json::ArrayIndex>(
                 m_queryOriginId)]);
 
-    BOX3D q(toBox3d(found["bounds"]));
+    try
+    {
+        BOX3D q(toBox3d(found["bounds"]));
 
-    // Clip the bounds to the queried origin bounds.  Don't just overwrite it -
-    // it's possible that both a bounds and an origin are specified.
-    m_queryBounds.clip(q);
+        // Clip the bounds to the queried origin bounds.  Don't just overwrite
+        // it - it's possible that both a bounds and an origin are specified.
+        m_queryBounds.clip(q);
 
-    log()->get(LogLevel::Debug) << "Query origin " <<
-        m_queryOriginId << ": " << found["id"].asString() <<
-        std::endl;
+        log()->get(LogLevel::Debug) << "Query origin " <<
+            m_queryOriginId << ": " << found["id"].asString() << std::endl;
+    }
+    catch (std::exception& e)
+    {
+        throwError(e.what());
+    }
 }
 
 QuickInfo EptReader::inspect()
 {
     QuickInfo qi;
 
-    initialize();
-
-    qi.m_bounds = toBox3d(m_info->json()["boundsConforming"]);
-    qi.m_srs = m_info->srs();
-    qi.m_pointCount = m_info->points();
-
-    for (Json::ArrayIndex i(0); i < m_info->schema().size(); ++i)
+    try
     {
-        qi.m_dimNames.push_back(m_info->schema()[i]["name"].asString());
+        initialize();
+
+        qi.m_bounds = toBox3d(m_info->json()["boundsConforming"]);
+        qi.m_srs = m_info->srs();
+        qi.m_pointCount = m_info->points();
+
+        for (Json::ArrayIndex i(0); i < m_info->schema().size(); ++i)
+        {
+            qi.m_dimNames.push_back(m_info->schema()[i]["name"].asString());
+        }
+    }
+    catch (std::exception& e)
+    {
+        throwError(e.what());
     }
 
     qi.m_valid = true;
@@ -317,90 +348,185 @@ void EptReader::addDimensions(PointLayoutPtr layout)
     m_xyzTransforms[0] = m_dimTypes[dimIndex(D::X)].m_xform;
     m_xyzTransforms[1] = m_dimTypes[dimIndex(D::Y)].m_xform;
     m_xyzTransforms[2] = m_dimTypes[dimIndex(D::Z)].m_xform;
+
+    try
+    {
+        for (const std::string dimName : m_args.addons().getMemberNames())
+        {
+            std::string root(m_args.addons()[dimName].asString());
+            const std::string postfix("ept-addon.json");
+            if (Utils::endsWith(root, postfix))
+            {
+                root = root.substr(0, root.size() - postfix.size());
+            }
+            root = arbiter::fs::expandTilde(root);
+
+            const arbiter::Endpoint ep(m_arbiter->getEndpoint(root));
+            const Json::Value addonInfo(parse(ep.get("ept-addon.json")));
+            const Dimension::Type type(getType(addonInfo));
+            const Dimension::Id id(layout->registerOrAssignDim(dimName, type));
+
+            m_addons.emplace_back(new Addon(*layout, ep, id));
+
+            log()->get(LogLevel::Debug) << "Registering addon dim " <<
+                dimName << ": " <<
+                Dimension::interpretationName(m_addons.back()->type()) <<
+                ", from " << root << std::endl;
+        }
+    }
+    catch (std::exception& e)
+    {
+        throwError(e.what());
+    }
 }
 
 void EptReader::ready(PointTableRef table)
 {
-    m_overlapKeys.clear();
-    m_overlapPoints = 0;
+    // These may not exist, in general they are only needed to track point
+    // origins and ordering for an EPT writer.
+    m_nodeIdDim = table.layout()->findDim("EptNodeId");
+    m_pointIdDim = table.layout()->findDim("EptPointId");
+
+    m_overlaps.clear();
 
     // Determine all overlapping data files we'll need to fetch.
-    overlaps();
-
-    log()->get(LogLevel::Debug) << "Overlap nodes: " << m_overlapKeys.size() <<
-        std::endl;
-    log()->get(LogLevel::Debug) << "Overlap points: " << m_overlapPoints <<
-        std::endl;
-
-    if (m_overlapPoints > 1e8)
+    try
     {
-        log()->get(LogLevel::Warning) << m_overlapPoints <<
+        overlaps();
+    }
+    catch (std::exception& e)
+    {
+        throwError(e.what());
+    }
+
+    uint64_t overlapPoints(0);
+    Json::Value json;
+
+    for (const auto& p : m_overlaps)
+    {
+        const Key& key(p.first);
+        const Json::UInt64 np(p.second);
+        overlapPoints += np;
+        json[key.toString()] = np;
+    }
+
+    log()->get(LogLevel::Debug) << "Overlap nodes: " << m_overlaps.size() <<
+        std::endl;
+    log()->get(LogLevel::Debug) << "Overlap points: " << overlapPoints <<
+        std::endl;
+
+    if (overlapPoints > 1e8)
+    {
+        log()->get(LogLevel::Warning) << overlapPoints <<
             " will be downloaded" << std::endl;
     }
+
+    if (m_nodeIdDim != Dimension::Id::Unknown)
+    {
+        // If we have an EPT writer in this pipeline, serialize some metadata
+        // which will be needed to create addon dimensions.
+        MetadataNode meta(table.privateMetadata("ept"));
+        meta.add("info", stringify(m_info->json()));
+        meta.add("keys", stringify(json));
+        meta.add("step", m_hierarchyStep);
+    }
 }
 
 void EptReader::overlaps()
 {
     // Determine all the keys that overlap the queried area by traversing the
-    // EPT hierarchy (see https://git.io/fAiuR).  Because this may require
-    // fetching lots of JSON files, it'll run in our thread pool.
-    const Key key(m_info->bounds());
-    const Json::Value hier(parse(m_ep->get(
-                    "ept-hierarchy/" + key.toString() + ".json")));
-    overlaps(hier, key);
-    m_pool->await();
+    // EPT hierarchy:
+    //      https://entwine.io/entwine-point-tile.html#ept-hierarchy)
+    //
+    // Because this may require fetching lots of JSON files, it'll run in our
+    // thread pool.
+    Key key;
+    key.b = m_info->bounds();
+    const std::string file("ept-hierarchy/" + key.toString() + ".json");
+
+    {
+        // First, determine the overlapping nodes from the EPT resource.
+        const Json::Value root(parse(m_ep->get(file)));
+        overlaps(*m_ep, m_overlaps, root, key);
+        m_pool->await();
+    }
+
+    for (auto& addon : m_addons)
+    {
+        // Next, determine the overlapping nodes from each addon dimension.
+        const Json::Value root(parse(addon->ep().get(file)));
+        overlaps(addon->ep(), addon->hierarchy(), root, key);
+        m_pool->await();
+    }
 }
 
-void EptReader::overlaps(const Json::Value& hier, const Key& key)
+void EptReader::overlaps(const arbiter::Endpoint& ep,
+        std::map<Key, uint64_t>& target, const Json::Value& hier,
+        const Key& key)
 {
     if (!key.b.overlaps(m_queryBounds)) return;
+    if (m_depthEnd && key.d >= m_depthEnd) return;
     const int64_t np(hier[key.toString()].asInt64());
     if (!np) return;
 
     if (np == -1)
     {
+        if (!m_hierarchyStep) m_hierarchyStep = key.d;
+
         // If the hierarchy points value here is -1, then we need to fetch the
         // hierarchy subtree corresponding to this root.
-        log()->get(LogLevel::Debug) << "Hierarchy: " << key.toString() <<
-            std::endl;
-
-        m_pool->add([this, key]()
+        m_pool->add([this, &ep, &target, key]()
         {
-            const auto subRoot(parse(m_ep->get(
+            const auto subRoot(parse(ep.get(
                             "ept-hierarchy/" + key.toString() + ".json")));
-            overlaps(subRoot, key);
+            overlaps(ep, target, subRoot, key);
         });
     }
     else
     {
         {
             std::lock_guard<std::mutex> lock(m_mutex);
-            m_overlapKeys.insert(key);
-            m_overlapPoints += static_cast<uint64_t>(np);
+            target[key] = static_cast<uint64_t>(np);
         }
 
         for (uint64_t dir(0); dir < 8; ++dir)
         {
-            overlaps(hier, key.bisect(dir));
+            overlaps(ep, target, hier, key.bisect(dir));
         }
     }
 }
 
 PointViewSet EptReader::run(PointViewPtr view)
 {
-    uint64_t i(0);
-    for (const Key& key : m_overlapKeys)
-    {
-        log()->get(LogLevel::Debug) << "Data " << ++i << "/" <<
-            m_overlapKeys.size() << ": " << key.toString() << std::endl;
+    // Start these at 1 to differentiate from points added by other stages,
+    // which will be ignored by the EPT writer.
+    uint64_t nodeId(1);
 
-        m_pool->add([this, &view, &key]()
+    for (const auto& entry : m_overlaps)
+    {
+        const Key& key(entry.first);
+
+        log()->get(LogLevel::Debug) << "Data " << nodeId << "/" <<
+            m_overlaps.size() << ": " << key.toString() << std::endl;
+
+        m_pool->add([this, &view, &key, nodeId]()
         {
+            uint64_t startId(0);
+
             if (m_info->dataType() == EptInfo::DataType::Laszip)
-                readLaszip(*view, key);
+                startId = readLaszip(*view, key, nodeId);
             else
-                readBinary(*view, key);
+                startId = readBinary(*view, key, nodeId);
+
+            // Read addon information after the native data, we'll possibly
+            // overwrite attributes.
+            for (const auto& addon : m_addons)
+            {
+                readAddon(*view, key, *addon, startId);
+            }
         });
+
+        ++nodeId;
     }
 
     m_pool->await();
@@ -412,7 +538,8 @@ PointViewSet EptReader::run(PointViewPtr view)
     return views;
 }
 
-void EptReader::readLaszip(PointView& dst, const Key& key) const
+uint64_t EptReader::readLaszip(PointView& dst, const Key& key,
+        const uint64_t nodeId) const
 {
     // If the file is remote (HTTP, S3, Dropbox, etc.), getLocalHandle will
     // download the file and `localPath` will return the location of the
@@ -431,18 +558,25 @@ void EptReader::readLaszip(PointView& dst, const Key& key) const
     std::lock_guard<std::mutex> lock(m_mutex);
     reader.prepare(table);
 
+    const uint64_t startId(dst.size());
+
+    uint64_t pointId(0);
     for (auto& src : reader.execute(table))
     {
         PointRef pr(*src);
         for (uint64_t i(0); i < src->size(); ++i)
         {
             pr.setPointId(i);
-            process(dst, pr);
+            process(dst, pr, nodeId, pointId);
+            ++pointId;
         }
     }
+
+    return startId;
 }
 
-void EptReader::readBinary(PointView& dst, const Key& key) const
+uint64_t EptReader::readBinary(PointView& dst, const Key& key,
+        const uint64_t nodeId) const
 {
     auto data(m_ep->getBinary("ept-data/" + key.toString() + ".bin"));
     ShallowPointTable table(*m_remoteLayout, data.data(), data.size());
@@ -450,18 +584,24 @@ void EptReader::readBinary(PointView& dst, const Key& key) const
 
     std::lock_guard<std::mutex> lock(m_mutex);
 
-    for (uint64_t i(0); i < table.numPoints(); ++i)
+    const uint64_t startId(dst.size());
+
+    uint64_t pointId(0);
+    for (uint64_t pointId(0); pointId < table.numPoints(); ++pointId)
     {
-        pr.setPointId(i);
-        process(dst, pr);
+        pr.setPointId(pointId);
+        process(dst, pr, nodeId, pointId);
     }
+
+    return startId;
 }
 
-void EptReader::process(PointView& dst, PointRef& pr) const
+void EptReader::process(PointView& dst, PointRef& pr, const uint64_t nodeId,
+        const uint64_t pointId) const
 {
     using D = Dimension::Id;
 
-    const point_count_t pointId(dst.size());
+    const point_count_t dstId(dst.size());
 
     const double x = pr.getFieldAs<double>(D::X) *
         m_xyzTransforms[0].m_scale.m_val + m_xyzTransforms[0].m_offset.m_val;
@@ -475,9 +615,9 @@ void EptReader::process(PointView& dst, PointRef& pr) const
 
     if (selected && m_queryBounds.contains(x, y, z))
     {
-        dst.setField(Dimension::Id::X, pointId, x);
-        dst.setField(Dimension::Id::Y, pointId, y);
-        dst.setField(Dimension::Id::Z, pointId, z);
+        dst.setField(Dimension::Id::X, dstId, x);
+        dst.setField(Dimension::Id::Y, dstId, y);
+        dst.setField(Dimension::Id::Z, dstId, z);
 
         for (const DimType& dt : m_dimTypes)
         {
@@ -486,27 +626,56 @@ void EptReader::process(PointView& dst, PointRef& pr) const
                 const double d = pr.getFieldAs<double>(dt.m_id) *
                     dt.m_xform.m_scale.m_val + dt.m_xform.m_offset.m_val;
 
-                dst.setField(dt.m_id, pointId, d);
+                dst.setField(dt.m_id, dstId, d);
             }
         }
+
+        dst.setField(m_nodeIdDim, dstId, nodeId);
+        dst.setField(m_pointIdDim, dstId, pointId);
     }
 }
 
-Json::Value EptReader::parse(const std::string& data) const
+void EptReader::readAddon(PointView& dst, const Key& key, const Addon& addon,
+        const uint64_t pointId) const
 {
-    Json::Value json;
-    Json::Reader reader;
-
-    if (!reader.parse(data, json, false))
+    const uint64_t np(addon.points(key));
+    if (!np)
     {
-        const std::string jsonError(reader.getFormattedErrorMessages());
-        if (!jsonError.empty())
+        // If our addon has no points, then we are reading a superset of this
+        // addon, in which case we need to zero-fill this dimension.
+        //
+        // For example, an EPT reader/writer loop may be performed with a
+        // spatially-bounded or resolution-limited query.  Then, this partial
+        // addon may be mapped to a well-known dimension, like Classification,
+        // for an EPT-read of the full dataset.  If the native EPT set already
+        // contains Classification, then we should overwrite it with zeroes
+        // where the addon leaves off.
+        for (uint64_t id(pointId); id < pointId + np; ++id)
         {
-            throwError("Error during parsing: " + jsonError);
+            dst.setField(addon.id(), id, 0);
         }
+
+        return;
     }
 
-    return json;
+    // If the addon hierarchy exists, it must match the EPT data.
+    if (np != m_overlaps.at(key)) throwError("Invalid addon hierarchy");
+
+    const auto data(addon.ep().getBinary(
+                "ept-data/" + key.toString() + ".bin"));
+    const uint64_t dimSize(Dimension::size(addon.type()));
+
+    if (np * dimSize != data.size())
+    {
+        throwError("Invalid addon content length");
+    }
+
+    const char* pos(data.data());
+    for (uint64_t id(pointId); id < pointId + np; ++id)
+    {
+        dst.setField(addon.id(), addon.type(), id, pos);
+        pos += dimSize;
+    }
 }
 
 } // namespace pdal
