@@ -384,6 +384,8 @@ void LasReader::extractHeaderMetadata(MetadataNode& forward, MetadataNode& m)
     addForwardMetadata(forward, m, "offset_z", m_header.offsetZ(),
         "The offset for Z values.", 20);
 
+    m.add("point_length", m_header.pointLen(),
+        "The size, in bytes, of each point records.");
     m.add("header_size", m_header.vlrOffset(),
         "The size, in bytes, of the header block, including any extension "
         "by specific software.");
@@ -483,13 +485,7 @@ void LasReader::readExtraBytesVlr()
 
 void LasReader::setSrs(MetadataNode& m)
 {
-    // If the user is already overriding this by setting it on the stage, we'll
-    // take their overridden value
-    SpatialReference srs = getSpatialReference();
-
-    if (srs.getWKT().empty())
-        srs = m_header.srs();
-    setSpatialReference(m, srs);
+    setSpatialReference(m, m_header.srs());
 }
 
 
