@@ -56,6 +56,8 @@
 #include <iostream>
 #include <vector>
 
+#include <json/json.h>
+
 #include <pdal/pdal_features.hpp>
 #include <pdal/DimUtil.hpp>
 #include <pdal/PDALUtils.hpp>
@@ -87,7 +89,8 @@ CREATE_STATIC_STAGE(LasWriter, s_info)
 std::string LasWriter::getName() const { return s_info.name; }
 
 LasWriter::LasWriter() : m_compressor(nullptr), m_ostream(NULL),
-    m_compression(LasCompression::None), m_srsCnt(0)
+    m_compression(LasCompression::None), m_srsCnt(0),
+    m_userVLRs(new Json::Value())
 {}
 
 
@@ -144,7 +147,7 @@ void LasWriter::addArgs(ProgramArgs& args)
     args.add("offset_x", "X offset", m_offsetX);
     args.add("offset_y", "Y offset", m_offsetY);
     args.add("offset_z", "Z offset", m_offsetZ);
-    args.add("vlrs", "List of VLRs to set", m_userVLRs);
+    args.add("vlrs", "List of VLRs to set", *m_userVLRs);
 }
 
 void LasWriter::initialize()
@@ -228,7 +231,7 @@ void LasWriter::prepared(PointTableRef table)
 // Capture user-specified VLRs
 void LasWriter::addUserVlrs()
 {
-    for (const auto& v : m_userVLRs)
+    for (const auto& v : *m_userVLRs)
     {
         uint16_t recordId(1);
         std::string userId("");
@@ -270,18 +273,15 @@ void LasWriter::fillForwardList()
 
     static const StringList offset = { "offset_x", "offset_y", "offset_z" };
 
-    static StringList all;
-    all.insert(all.begin(), header.begin(), header.end());
-    all.insert(all.begin(), scale.begin(), scale.end());
-    all.insert(all.begin(), offset.begin(), offset.end());
-
     // Build the forward list, replacing special keywords with the proper
     // field names.
     for (auto& name : m_forwardSpec)
     {
         if (name == "all")
         {
-            m_forwards.insert(all.begin(), all.end());
+            m_forwards.insert(header.begin(), header.end());
+            m_forwards.insert(scale.begin(), scale.end());
+            m_forwards.insert(offset.begin(), offset.end());
             m_forwardVlrs = true;
         }
         else if (name == "header")
@@ -294,7 +294,11 @@ void LasWriter::fillForwardList()
             m_forwards.insert("dataformat_id");
         else if (name == "vlr")
             m_forwardVlrs = true;
-        else if (Utils::contains(all, name))
+        else if (
+            Utils::contains(header, name) ||
+            Utils::contains(scale, name) ||
+            Utils::contains(offset, name)
+        )
             m_forwards.insert(name);
         else
             throwError("Error in 'forward' option.  Unknown field for "
@@ -822,11 +826,16 @@ bool LasWriter::processPoint(PointRef& point)
 }
 
 
+void LasWriter::prerunFile(const PointViewSet& pvSet)
+{
+    m_scaling.setAutoXForm(pvSet);
+}
+
+
 void LasWriter::writeView(const PointViewPtr view)
 {
     Utils::writeProgress(m_progressFd, "READYVIEW",
         std::to_string(view->size()));
-    m_scaling.setAutoXForm(view);
 
     point_count_t pointLen = m_lasHeader.pointLen();
 
@@ -933,35 +942,37 @@ bool LasWriter::writeLasZipBuf(PointRef& point)
     p.intensity = point.getFieldAs<uint16_t>(Id::Intensity);
     p.scan_direction_flag = scanDirectionFlag;
     p.edge_of_flight_line = edgeOfFlightLine;
+    p.synthetic_flag = classFlags & 0x1;
+    p.keypoint_flag = (classFlags >> 1) & 0x1;
+    p.withheld_flag = (classFlags >> 2) & 0x1;
+    p.user_data = point.getFieldAs<uint8_t>(Id::UserData);
+    p.point_source_ID = point.getFieldAs<uint16_t>(Id::PointSourceId);
 
     if (has14Format)
     {
-        p.extended_point_type = 1;
+        p.classification = (classification & 0x1F) | (classFlags << 5);
+        p.scan_angle_rank = point.getFieldAs<int8_t>(Id::ScanAngleRank);
+        p.number_of_returns = (std::min)((uint8_t)7, numberOfReturns);
+        p.return_number = (std::min)((uint8_t)7, returnNumber);
 
-        p.extended_return_number = returnNumber;
-        p.extended_number_of_returns = numberOfReturns;
-        p.extended_scanner_channel = scanChannel;
         // This should always work if ScanAngleRank isn't wonky.
         p.extended_scan_angle = static_cast<laszip_I16>(
             std::round(point.getFieldAs<float>(Id::ScanAngleRank) / .006f));
+        p.extended_point_type = 1;
+        p.extended_scanner_channel = scanChannel;
         p.extended_classification_flags = classFlags;
         p.extended_classification = classification;
-        p.classification = (classification & 0x1F) | (classFlags << 5);
-//        p.scan_angle_rank = point.getFieldAs<int8_t>(Id::ScanAngleRank);
+        p.extended_return_number = returnNumber;
+        p.extended_number_of_returns = numberOfReturns;
     }
     else
     {
-        p.synthetic_flag = classFlags & 0x1;
-        p.keypoint_flag = (classFlags >> 1) & 0x1;
-        p.withheld_flag = (classFlags >> 2) & 0x1;
         p.return_number = returnNumber;
         p.number_of_returns = numberOfReturns;
         p.scan_angle_rank = point.getFieldAs<int8_t>(Id::ScanAngleRank);
         p.classification = classification;
+        p.extended_point_type = 0;
     }
-    p.user_data = point.getFieldAs<uint8_t>(Id::UserData);
-
-    p.point_source_ID = point.getFieldAs<uint16_t>(Id::PointSourceId);
 
     if (m_lasHeader.hasTime())
         p.gps_time = point.getFieldAs<double>(Id::GpsTime);
