@@ -1,4 +1,4 @@
-#!/bin/sh -e
+#!/bin/ash
 # Builds and tests PDAL
 
 gcc --version
@@ -10,11 +10,8 @@ mkdir -p _build || exit 1
 cd _build || exit 1
 
 cmake .. \
-    -G "Unix Makefiles" \
+    -G Ninja \
     -DCMAKE_BUILD_TYPE=Release \
-    -DCMAKE_C_COMPILER=gcc \
-    -DCMAKE_CXX_COMPILER=g++ \
-    -DCMAKE_MAKE_PROGRAM=make \
     -DBUILD_PLUGIN_PYTHON=ON \
     -DBUILD_PLUGIN_CPD=ON \
     -DBUILD_PLUGIN_GREYHOUND=ON \
@@ -28,26 +25,24 @@ cmake .. \
     -DWITH_LAZPERF=ON \
     -DWITH_TESTS=ON
 
-make -j2
-LD_LIBRARY_PATH=./lib
+ninja -v
 ctest -V
-make install
+ninja install
+cd /
 
 # Python extension testing
-pip install packaging
+pip3 install packaging cython
 git clone https://github.com/PDAL/python.git pdal-python
 cd pdal-python
 git checkout 2.0.0
-python setup.py build
-echo "current path: " `pwd`
-export PDAL_TEST_DIR=/pdal/_build/test
-python setup.py test
+python3 setup.py build
+PDAL_TEST_DIR=/pdal/_build/test python3 setup.py test
 
 for EXAMPLE in writing writing-filter writing-kernel writing-reader writing-writer
 do
     cd /pdal/examples/$EXAMPLE
     mkdir -p _build || exit 1
     cd _build || exit 1
-    cmake -G "Unix Makefiles" .. && \
-    make
+    cmake -G "Ninja" .. && \
+    ninja
 done
