@@ -318,6 +318,70 @@ TEST(SpatialReferenceTest, test_vertical_and_horizontal)
 
 }
 
+TEST(SpatialReferenceTest, readerOptions)
+{
+    const SpatialReference utm16("EPSG:26916");
+    SpatialReference native;
+
+    {
+        Options o;
+        o.add("filename", Support::datapath("las/test_utm17.las"));
+        LasReader r;
+        r.setOptions(o);
+        const auto qi(r.preview());
+        native = qi.m_srs;
+    }
+
+    {
+        Options o;
+        o.add("filename", Support::datapath("las/test_utm17.las"));
+        o.add("spatialreference", "EPSG:26916");
+        LasReader r;
+        r.setOptions(o);
+
+        PointTable t;
+        r.prepare(t);
+        EXPECT_EQ(r.getSpatialReference(), utm16);
+    }
+
+    {
+        Options o;
+        o.add("filename", Support::datapath("las/test_utm17.las"));
+        o.add("override_srs", "EPSG:26916");
+        LasReader r;
+        r.setOptions(o);
+
+        PointTable t;
+        r.prepare(t);
+        EXPECT_EQ(r.getSpatialReference(), utm16);
+    }
+
+    {
+        Options o;
+        o.add("filename", Support::datapath("las/test_utm17.las"));
+        o.add("default_srs", "EPSG:26916");
+        LasReader r;
+        r.setOptions(o);
+
+        PointTable t;
+        r.prepare(t);
+        EXPECT_EQ(r.getSpatialReference(), native);
+    }
+
+    {
+        Options o;
+        // This file has no spatial reference.
+        o.add("filename", Support::datapath("las/100-points.las"));
+        o.add("default_srs", "EPSG:26916");
+        LasReader r;
+        r.setOptions(o);
+
+        PointTable t;
+        r.prepare(t);
+        EXPECT_EQ(r.getSpatialReference(), SpatialReference("EPSG:26916"));
+    }
+}
+
 TEST(SpatialReferenceTest, merge)
 {
     Options o1;
@@ -384,9 +448,9 @@ TEST(SpatialReferenceTest, test_bounds)
     BOX2D box17(289814.15, 4320978.61, 289818.50, 4320980.59);
     pdal::Polygon p(box17);
     p.setSpatialReference(utm17);
-    pdal::Polygon p2 = p.transform(wgs84);
+    p.transform(wgs84);
 
-    BOX3D b2 = p2.bounds();
+    BOX3D b2 = p.bounds();
     EXPECT_FLOAT_EQ(static_cast<float>(b2.minx), -83.42759776f);
     EXPECT_FLOAT_EQ(static_cast<float>(b2.miny), 39.01259905f);
     EXPECT_FLOAT_EQ(static_cast<float>(b2.maxx), -83.427551f);
@@ -414,6 +478,8 @@ TEST(SpatialReferenceTest, issue_1989)
     EXPECT_EQ(-32, south.getUTMZone());
 }
 
+// Ilvis needs XML2
+#ifdef PDAL_HAVE_LIBXML2
 TEST(SpatialReferenceTest, set_srs)
 {
     StageFactory factory;
@@ -444,6 +510,8 @@ TEST(SpatialReferenceTest, set_srs)
     EXPECT_NE(m.value().find("AUTHORITY[\"EPSG\",\"4326\"]]"),
         std::string::npos);
 
+    //
+    s = factory.createStage("readers.ilvis2");
     ops2.add("spatialreference", "EPSG:2029");
     s->setOptions(ops2);
     s->prepare(t);
@@ -451,5 +519,6 @@ TEST(SpatialReferenceTest, set_srs)
     EXPECT_NE(m.value().find("AUTHORITY[\"EPSG\",\"2029\"]]"),
         std::string::npos);
 }
+#endif // PDAL_HAVE_LIBXML2
 
 } // namespace pdal
